/*	License
 * 		inside EU: EUPL
 * 		outside EU: GPL or EUPL
 *  The program comes AS IS and without any warranty unless required by law
 *  (C) 2012 thecodemonkey
 */

package scheduler2;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;


public class Scheduler2 {
	protected List<SchedulerRunnable2> queued;
	protected List<SchedulerRunnable2> stopped;
	protected Map<Integer, SchedulerRunnable2> active;
	
	protected volatile boolean running;
	protected volatile int maxThreads, threadCounter;//, taskCounter, numberOfTasks;
	protected SchedulerClient client;
	
	public Scheduler2(SchedulerClient client, int maxThreads) {
		this.client=client;
		queued=Collections.synchronizedList(new LinkedList<SchedulerRunnable2>());
		stopped=Collections.synchronizedList(new LinkedList<SchedulerRunnable2>());
		active=Collections.synchronizedMap(new HashMap<Integer,SchedulerRunnable2>());
		this.maxThreads=maxThreads;
	}
	
	public void setMaxThreads(int maxThreads) {
		this.maxThreads = maxThreads;
	}
	
	public void start() {
		running=true;
		System.out.println("start");
		int k=Math.min(queued.size(), maxThreads);
		
		for (int i=0;i<k;i++) {
			startTask();
		}
	}

	public int addTaskStopped(SchedulerRunnable2 t) {
		synchronized(stopped){
			stopped.add(t);
		}
		return t.getId();
	}
	
	public int addTask(SchedulerRunnable2 t) {
		synchronized(queued) {
			queued.add(t);
			if (running) {
				if (threadCounter<maxThreads) {
					startTask();
				}
			}
		}
		return t.getId();
	}
	
	public void stopAll() {
		synchronized (stopped) {
			synchronized (active) {
				running=false;
				Iterator<SchedulerRunnable2>it =active.values().iterator();
				while (it.hasNext()) {
					SchedulerRunnable2 t=it.next();
					t.stop();
					stopped.add(t);
				}
				active.clear();
			}
		
			synchronized (queued) {
				Iterator<SchedulerRunnable2>it =queued.iterator();
				while (it.hasNext()) {
					stopped.add(it.next());
				}
				queued.clear();
			}
		}
	}
	
	public SchedulerRunnable2 remove(int id) {
		synchronized (active) {
			SchedulerRunnable2 t=active.get(id);
			
			if (t!=null) {
				t.stop();
				SchedulerRunnable2 thread = active.remove(id);
				synchronized (queued) {
					if (!queued.isEmpty()) {
						startTask();
					} else {
						synchronized (stopped) {
							if (active.isEmpty() &&stopped.isEmpty()) {
								running = false;
								client.allDone();
							}
						}
						
					}
						
				}
				
				return thread;
			} else {
				synchronized (queued) {
					for (SchedulerRunnable2 thread:queued) {
						if (thread.id==id) {
							queued.remove(thread);
							return thread;
						}
					}
				}
				synchronized (stopped) {
					for (SchedulerRunnable2 thread:stopped) {
						if (thread.id==id) {
							stopped.remove(thread);
							return thread;
						}
					}
				}
				
				return null;
			}
		}
		
	}
	
	public void removeAll() {
		stopAll();
		queued.clear();
		stopped.clear();
	}
	
	public SchedulerRunnable2 stop(int id) {
		synchronized (active){
			SchedulerRunnable2 t=active.remove(id);
			
			if (t!=null) {
				t.stop();
				stopped.add(t);
				if (active.isEmpty() && queued.isEmpty()) running=false;
			} else {
				synchronized (queued) {
					for (SchedulerRunnable2 thread:queued) {
						if (thread.id==id) {
							queued.remove(thread);
							synchronized (stopped) {
								stopped.add(thread);
							}
							return thread;
						}
					}
				}
			}
			return t;
		}
	}
	
	protected void startTask() {
		if (running) {
			SchedulerRunnable2 thread=queued.remove(0);
			active.put(thread.id,  thread);
			threadCounter++; 
			new SchedulerThread(thread).start();
		}
	}
	
	protected class SchedulerThread extends Thread {
		
		private SchedulerRunnable2 concreteTask;
		
		public SchedulerThread(SchedulerRunnable2 concreteTask) {
			this.concreteTask=concreteTask;
		}
		
		public SchedulerRunnable2 getConcreteTask() {
			return concreteTask;
		}
		
		@Override
		public void run() {
			concreteTask.run();
			threadCounter--;
			synchronized (active) {
				active.remove(concreteTask.getId());
			}
			synchronized (queued) {
				if (!queued.isEmpty()) {
					startTask();
				} else {
					synchronized (active) {
						synchronized (stopped) {
							if (active.isEmpty() &&stopped.isEmpty()) {
								running = false;
								client.allDone();
							} 
						}
					}
					
				}
			}
			
		}
	}
	
	public boolean isRunning() {
		synchronized (active) {
			return running;
		}
	}
	

	public Collection<SchedulerRunnable2> getActive() {
		synchronized (active) {
			Collection<SchedulerRunnable2> threads=active.values();
			List<SchedulerRunnable2> list=new LinkedList<SchedulerRunnable2>();
			for (SchedulerRunnable2 t:threads) {
				list.add(t);
			}
			return list;
		}
		
	}
	
	public Collection<SchedulerRunnable2> getUnfinished() {
		synchronized (active) {
			Collection<SchedulerRunnable2> threads=active.values();
			List<SchedulerRunnable2> list=new LinkedList<SchedulerRunnable2>();
			for (SchedulerRunnable2 t:threads) {
				list.add(t);
			}
			synchronized (queued) {
				Iterator<SchedulerRunnable2> it=queued.iterator();
				while (it.hasNext()) {
					list.add(it.next());
				}
			}
			
			synchronized (queued) {
				Iterator<SchedulerRunnable2> it=stopped.iterator();
				while (it.hasNext()) {
					list.add(it.next());
				}
			}
			
			return list;
		}
	
	}

	public boolean isRunning(int id) {
		synchronized (active) {
			if (active.containsKey(id)) {
				return true;
			} else {
				synchronized (queued) {
					for (SchedulerRunnable2 sr:queued) {
						if (sr.getId()==id) return true;
					}
				}
				return false;
			}
		}
	}
	
	public SchedulerRunnable2 enqueueStopped(int id) {
		synchronized (stopped) {
			Iterator<SchedulerRunnable2> it=stopped.iterator();
			SchedulerRunnable2 sr = null;
			while (it.hasNext()) {
				SchedulerRunnable2 task=it.next();
				if (task.getId()==id) {
					synchronized (queued) {
						queued.add(0,task);
					}
					it.remove();
					sr=task;
					break;
				}
			}
			return sr;
		}
	}
	
	public void enqueueAllStopped() {
		synchronized (stopped) {
			synchronized (queued) {
				int i=0;
				while (!stopped.isEmpty()) {
					queued.add(i++, stopped.remove(0));
				}
			}
		}
	}
	
	public SchedulerRunnable2 getStoppedById(int id) {
		synchronized (stopped) {
			for (SchedulerRunnable2 r:stopped) {
				if (r.id == id) return r;
			}
			return null;
		}
	}
	
	public Collection<SchedulerRunnable2> getStopped() {
		List<SchedulerRunnable2> list=new LinkedList<>();
		synchronized (stopped) {
			for (SchedulerRunnable2 r:stopped) {
				list.add(r);
			}
			return list;
		}
	}
	
	public void reEnqueue(int id) {
		synchronized (active) {
			synchronized (queued) {
				if (active.containsKey(id)) {
					queued.add(active.remove(id));
				}
			}
		}
		
	}
}
